import click
import http.client


@click.command(
    help="Select based on id or name. Depends on current selection. "
    "If nothing is selected, it will try to select a course, "
    "then an exercise-series, then an exercise. "
    "Will not work with invalid id's or names without flag."
)
@click.option(
    "--hidden", "-hidden",
    help="Access a hidden series. Only available when trying to select "
    "exercise series."
    "\n\nUsage: dodona select --hidden <TOKEN> <SERIES_ID>"
    )
@click.option(
    "--other", "-other",
    help="Select a course that you're not registered for. Only works with "
    "an id, not a name.",
    is_flag=True, default=False
)
@click.argument('thing')
def select(thing, hidden, other):
    from dodonacli.source import (
        set_data, get_data, pretty_print, pretty_printer
    )

    # Read configs in
    config = get_data.get_configs()
    settings = get_data.get_settings()

    # Start up the connection to Dodona
    connection = http.client.HTTPSConnection("dodona.be")
    headers = {
        "Content-type": "application/json",
        "Accept": "application/json",
        "Authorization": config['TOKEN']
    }

    if config['course_id'] is None:
        if settings['display_series_after_select']:
            config = select_course(
                connection, headers, thing, config,
                {
                    'new_lines_above': settings['new_lines_above'],
                    'new_lines_below': 1
                },
                other
            )
            # Print available series
            json_data = get_data.series_data(
                connection, headers, config['course_id']
            )
            pretty_print.print_series_data(
                json_data, {'new_lines_below': settings['new_lines_below']}
            )
        else:
            config = select_course(
                connection, headers, thing, config, settings, other
            )

    elif config['serie_id'] is None:
        if settings['display_exercises_after_select']:
            if hidden:
                config = select_hidden_series(
                    connection, headers, thing, hidden, config,
                    {
                        'new_lines_above': settings['new_lines_above'],
                        'new_lines_below': 1
                    }
                )
            else:
                config = select_series(
                    connection, headers, thing, config,
                    {
                        'new_lines_above': settings['new_lines_above'],
                        'new_lines_below': 1
                    }
                )
            # Print available exercises
            if config['serie_token'] is None:
                serie_token = ""
            else:
                serie_token = "?token=" + config['serie_token']

            json_data = get_data.exercises_data(
                connection, headers, config['serie_id'], serie_token
            )
            pretty_print.print_exercise_data(
                json_data, {'new_lines_below': settings['new_lines_below']}
            )
        else:
            if hidden:
                config = select_hidden_series(
                    connection, headers, thing, hidden, config, settings
                )
            else:
                config = select_series(
                    connection, headers, thing, config, settings
                )

    elif config['exercise_id'] is None:
        if settings['display_exercise_after_select']:
            config = select_exercise(
                connection, headers, thing, config,
                {
                    'new_lines_above': settings['new_lines_above'],
                    'new_lines_below': 1
                }
            )
            # Print exercise-description
            json_data = get_data.exercise_data(
                connection, headers, config['course_id'], config['exercise_id']
            )
            pretty_print.print_exercise(
                json_data, config['TOKEN'],
                {'new_lines_below': settings['new_lines_below']}
            )
        else:
            config = select_exercise(
                connection, headers, thing, config, settings
            )

    else:
        # You can't select more when everything is already selected
        pretty_printer.custom_print(
            "There is already an exercise selected.\nPlease remove selection "
            "with 'dodona up' before selecting a new exercise.",
            settings
        )
        return

    # Save selections in config file
    set_data.dump_config(config)

    return


def select_course(
    connection: http.client.HTTPSConnection, headers: dict, thing: str,
    config: dict, settings: dict, other: bool
) -> dict:
    if not other:
        return select_registered_course(
            connection, headers, thing, config, settings
        )
    else:
        return select_unregistered_course(
            connection, headers, thing, config, settings
        )


def select_registered_course(
    connection: http.client.HTTPSConnection,
    headers: dict, course: str, config: dict, settings: dict
) -> dict:
    from dodonacli.source import pretty_printer, get_data

    # Get all registered courses to check if a valid course was selected
    data_courses = get_data.courses_data(connection, headers)
    courses = {str(course['id']): course['name'] for course in data_courses}

    # First try to match on id, else on name
    if course.isnumeric() and course in courses:
        config['course_id'] = course
        config['course_name'] = courses[course]

        pretty_printer.custom_print(
            f"Course [bold]\"{courses[course]}\"[/] selected.",
            settings, pretty=True
        )
    else:
        for course_item in courses.items():
            if course.lower() in course_item[1].lower():
                config['course_id'] = course_item[0]
                config['course_name'] = course_item[1]

                pretty_printer.custom_print(
                    f"Course [bold]\"{courses[course_item[0]]}\"[/] selected.",
                    settings, pretty=True
                )

                # We found it, we can stop searching
                break

    # If config_id is still empty, nothing is found:
    if config['course_id'] is None:
        pretty_printer.custom_print(
           "Not a valid course-id or -name.", settings
        )
        exit(1)

    return config


def select_unregistered_course(
    connection: http.client.HTTPSConnection,
    headers: dict, course_id: str, config: dict, settings: dict
) -> dict:
    import json
    from dodonacli.source import pretty_printer, get_data

    # Has to be an id, because it's not possible to match against names
    if not course_id.isnumeric():
        pretty_printer.custom_print(
            "The selection needs to be an id (all numbers). "
            "It's not possible to match by name.", settings
        )
        exit(1)

    link = f"/courses/{course_id}"
    connection = get_data.handle_connection_request(
        connection, "GET", link, headers
    )

    res = connection.getresponse()

    if res.status != 200:
        pretty_printer.custom_print(
            "Something went wrong trying to select this course.\n"
            "Make sure you have a valid course-id.\n"
            "\tResponse code: " + str(res.status) +
            "\n\tReason: " + str(res.reason).strip(),
            settings, pretty=True
        )
        exit(1)

    json_data = json.loads(res.read())
    config['course_id'] = course_id
    config['course_name'] = json_data['name']

    pretty_printer.custom_print(
        f"Course [bold]\"{json_data['name']}\"[/] selected.",
        settings, pretty=True
    )

    return config


def select_series(connection: http.client.HTTPSConnection, headers: dict,
                  thing: str, config: dict, settings: dict):
    from dodonacli.source import pretty_printer, get_data

    data_series = get_data.series_data(
        connection, headers, config['course_id']
    )

    series = {str(serie['id']): serie['name'] for serie in data_series}

    if thing.isnumeric() and thing in series:
        config['serie_id'] = thing
        config['serie_name'] = series[thing]
        pretty_printer.custom_print(
            f"Series [bold]\"{series[thing]}\"[/] selected.",
            settings, pretty=True
        )
    else:
        for serie in series.items():
            if thing.lower() in serie[1].lower():
                config['serie_id'] = serie[0]
                config['serie_name'] = serie[1]
                pretty_printer.custom_print(
                    f"Series [bold]\"{series[serie[0]]}\"[/] selected.",
                    settings, pretty=True
                )
                break
    if config['serie_id'] is None:
        pretty_printer.custom_print(
            "Not a valid series id or -name!", settings
        )
        exit(1)

    return config


def select_hidden_series(
    connection: http.client.HTTPSConnection, headers: dict,
    series_id: str, series_token: str, config: dict, settings: dict
):
    import json
    from dodonacli.source import pretty_printer, get_data

    if not series_id.isnumeric():
        pretty_printer.custom_print(
            "Well, this won't work without a valid series-ID (only numeric "
            "characters). Please try again.",
            settings
        )
        exit(1)

    link = f"/series/{series_id}?token={series_token}"
    connection = get_data.handle_connection_request(
        connection, "GET", link, headers)

    res = connection.getresponse()

    if res.status != 200:
        pretty_printer.custom_print(
            "Something went wrong trying to select the hidden series.\n"
            "Check if you got the right series-id "
            "(seen in /series/<serie_id> in the link you got).\n"
            "\tResponse code: " + str(res.status) +
            "\n\tReason: " + str(res.reason).strip(),
            settings
        )
        exit(1)

    json_data = json.loads(res.read())
    config['serie_id'] = series_id
    config['serie_name'] = json_data['name']
    config['serie_token'] = series_token

    pretty_printer.custom_print(
        f"Series [bold]\"{json_data['name']}\"[/] selected.",
        settings, pretty=True
    )

    return config


def select_exercise(connection: http.client.HTTPSConnection, headers: dict,
                    selection: str, config: dict, settings: dict) -> dict:
    import textwrap, os
    from dodonacli.source import pretty_console, get_data, save_boilerplate

    # Token for hidden series
    if config['serie_token'] is None:
        serie_token = ""
    else:
        serie_token = "?token=" + config['serie_token']

    data_exercises = get_data.exercises_data(
        connection, headers, config['serie_id'], serie_token
    )

    selected_exercise = {}

    user_input_is_numeric = selection.isnumeric()
    if user_input_is_numeric:
        selection = int(selection)
    else:
        selection = selection.lower()

    for exercise in data_exercises:
        if user_input_is_numeric:
            if exercise['id'] == selection:
                selected_exercise = exercise
                break
        elif selection in exercise['name'].lower():
            selected_exercise = exercise
            break

    if selected_exercise == {}:
        print(
            '\n' * settings['new_lines_above']
            + "Not a valid exercise-id or -name."
            + '\n' * settings['new_lines_below']
        )
        return config

    # Store selection
    config['exercise_id'] = str(selected_exercise['id'])
    config['exercise_name'] = selected_exercise['name']
    # Programming language may not exist (f.e. for ContentPage)
    programming_language = selected_exercise.get('programming_language')
    if programming_language:
        config['programming_language'] = programming_language['name']

    pretty_console.console.print(
        '\n' * settings['new_lines_above']
        + f"Exercise [bold]{selected_exercise['name']}[/] selected."
    )

    boilerplate = selected_exercise.get('boilerplate')
    if boilerplate and boilerplate.strip() != "":
        save_boilerplate.save_boilerplate(
            boilerplate,
            selected_exercise['name'],
            programming_language['extension']
        )

    print('\n' * settings['new_lines_below'], end='')

    return config



import click


# Function and file have to be called cli_next, as next
# is a Python built-in. The command, however, will still be called
# with 'next', as click provides the 'name=' argument
@click.command(
    name="next",
    help="WARNING: might overwrite 'boilerplate' file! "
    "Move to the next type of what you have selected. "
    "It loops around to the beginning if the current selection "
    "is at the end of the 'list'. If some boilerplate is attached "
    "to the next exercise, it will put that in a file."
)
@click.option(
    "-r", "--reverse",
    help="Goes to the previous instead of the next.",
    is_flag=True, default=False
)
@click.option(
    "-u", "--unsolved",
    help="Find the next unsolved item. "
    "Currently only available for exercises, not series or courses.",
    is_flag=True, default=False
)
def cli_next(reverse, unsolved):
    import http.client
    from dodonacli.source import get_data, set_data, pretty_printer

    # Read configs in
    config = get_data.get_configs()
    settings = get_data.get_settings()

    # Start up the connection to Dodona
    connection = http.client.HTTPSConnection("dodona.be")
    headers = {
        "Content-type": "application/json",
        "Accept": "application/json",
        "Authorization": config['TOKEN']
    }

    if config.get('exercise_id') is not None:
        config = get_next_exercise(
            config, settings, connection, headers, reverse, unsolved
        )

    elif config.get('serie_id') is not None:
        config = get_next_series(
            config, settings, connection, headers, reverse, unsolved
        )

    elif config.get('course_id') is not None:
        config = get_next_course(
            config, settings, connection, headers, reverse, unsolved
        )

    else:
        pretty_printer.custom_print(
            "Can't select a next course when non are selected.",
            settings
        )
        return

    set_data.dump_config(config)


def get_next_exercise(config, settings, connection, headers, reverse, unsolved):
    from dodonacli.source import (
        get_data, pretty_print, pretty_printer, save_boilerplate
    )

    # Get all exercises of selected series
    exercise_data_json = get_data.exercises_data(
        connection, headers, config['serie_id'],
        f"/?token={config['serie_token']}" if config['serie_token'] else ""
    )

    # Calculate some needed values
    id_list = [exercise['id'] for exercise in exercise_data_json]
    previous_id = config['exercise_id']
    previous_id_index = id_list.index(int(previous_id))

    exercises_dict = {
        exercise['id']: {
            'name': exercise['name'],
            'boilerplate': exercise.get('boilerplate') or "",
            'programming_language': exercise.get('programming_language') or None,
            'accepted': exercise.get('accepted') or exercise.get('has_read'),
            'description_url': exercise['description_url']
        } for exercise in exercise_data_json
    }

    # Find the next exercise (loop back to front if it was the last)
    next_id = -1
    if not unsolved:
        next_id = id_list[(previous_id_index + 1 - (2 * reverse)) % len(id_list)]
    else:
        i = 1
        while next_id == -1 and i < len(id_list):
            index = (previous_id_index + i * (-1) ** reverse) % len(id_list)
            key = id_list[index]
            if not exercises_dict[key]['accepted']:
                next_id = id_list[index]
            i += 1
        if next_id == -1:
            pretty_printer.custom_print(
                "You already solved everything, there is no unsolved "
                "exercise to go to!"
            )
            return config

    # Store new exercise
    config['exercise_id'] = str(next_id)
    config['exercise_name'] = exercises_dict[next_id]['name']
    if exercises_dict[next_id]['programming_language']:
        config['programming_language'] = (
            exercises_dict[next_id]['programming_language']['name']
        )

    prefixes = make_visual_representation(
        previous_id, previous_id_index, next_id, id_list
    )

    pretty_print.print_exercise_data(exercise_data_json, settings, prefixes)

    # Handle potential boilerplate.
    boilerplate = exercises_dict[next_id]['boilerplate']
    if boilerplate is not None and boilerplate.strip() != "":
        save_boilerplate.save_boilerplate(
            boilerplate,
            exercises_dict[next_id]['name'],
            exercises_dict[next_id]['programming_language']['extension']
        )

    return config


def get_next_series(config, settings, connection, headers, reverse, unsolved):
    if config['serie_token']:
        print(
            '\n' * settings['new_lines_below']
            + "Sorry, you can't use 'next' when inside a hidden series."
            + '\n' * settings['new_lines_below']
        )
        return config

    # Only import when we need it, speeds things up
    from dodonacli.source import get_data, pretty_print

    # Get all series of selected course
    series_data_json = get_data.series_data(
        connection, headers, config['course_id']
    )

    # Take only necessary info from the large json
    # Series have an order, so sort them so the order is guarenteed
    series_list = [
        {'id': series['id'], 'name': series['name'], 'order': series['order']}
        for series in series_data_json
    ]
    series_list.sort(key=lambda x: x['order'])

    # Using the sorted list here to keep ensuring the same order
    id_list = [series['id'] for series in series_list]
    previous_id = config['serie_id']
    previous_id_index = id_list.index(int(previous_id))

    # Find the next series (loop back to front if it was the last)
    # If series ever get a solved/unsolved status support in API, this can get
    # the same logic found in get_next_exercise()
    if unsolved:
        print(
            '\n' * settings['new_lines_above']
            + "Unsolved flag not supported yet for series and courses."
            + '\n' * settings['new_lines_below']
        )
    next_id = id_list[(previous_id_index + 1 - (2 * reverse)) % len(id_list)]

    # Store new series
    config['serie_id'] = str(next_id)
    config['serie_name'] = [
        series['name']
        for series in series_list
        if series['id'] == next_id
    ][0]

    prefixes = make_visual_representation(
        previous_id, previous_id_index, next_id, id_list
    )
    pretty_print.print_series_data(series_data_json, settings, prefixes=prefixes)

    return config


def get_next_course(config, settings, connection, headers, reverse, unsolved):
    from dodonacli.source import get_data, pretty_print, pretty_printer

    # Get all registred courses
    course_data_json = get_data.courses_data(connection, headers)

    id_list = [course['id'] for course in course_data_json]
    previous_id = config['course_id']
    previous_id_index = id_list.index(int(previous_id))

    # Find the next course (loop back to front if it was the last)
    # If courses get more data that indicates if it's completely solved,
    # then this will get the same logic found in get_next_exercise()
    if unsolved:
        pretty_printer.custom_print(
            "Unsolved flag not supported yet for series and courses.",
            {'new_lines_above': settings['new_lines_above']}
        )

    next_id = id_list[(previous_id_index + 1 - (2 * reverse)) % len(id_list)]

    # Store new course
    config['course_id'] = str(next_id)
    config['course_name'] = [
        course['name']
        for course in course_data_json
        if course['id'] == next_id
    ][0]

    prefixes = make_visual_representation(
        previous_id, previous_id_index, next_id, id_list
    )
    pretty_print.print_courses_data(
        course_data_json,
        {'new_lines_below': settings['new_lines_below']},
        prefixes=prefixes
    )

    return config


def make_visual_representation(
    previous_id, previous_id_index, next_id, id_list
) -> dict:
    next_id_index = id_list.index(int(next_id))

    # Visual arrow representation of the jump:
    prefixes = {}
    if next_id_index > previous_id_index:
        prefixes[str(next_id)] = "   \u2570\u2B9E\t"
        prefixes[str(previous_id)] = "   \u256D\u2500\t"
        for e in id_list[previous_id_index + 1:next_id_index]:
            prefixes[str(e)] = "   \u2502\t"
    else:
        prefixes[str(next_id)] = "   \u256D\u2B9E\t"
        prefixes[str(previous_id)] = "   \u2570\u2500\t"
        for e in id_list[next_id_index + 1:previous_id_index]:
            prefixes[str(e)] = "   \u2502\t"

    return prefixes



from . import pretty_console


def custom_print(text: str, settings: dict, pretty=False, end='\n'):
    """
    Prints out the text with the amount of newlines specified in settings
    :param text: text to print
    :param settings: dict with settings
    :param pretty: whether to use the pretty_console print, or the standard
                   Python print
    """
    if pretty:
        printer = pretty_console.console.print
    else:
        printer = print

    printer(
        '\n' * settings.get('new_lines_above', 0)
        + text
        + '\n' * settings.get('new_lines_below', 0),
        end=end
    )



"""
Global console object to be able to use it anywhere in the code by just
importing it.
>>> from pretty_console import console
"""
from rich.console import Console
console = Console(color_system="standard")



import click


@click.command(
    help="Display info based on the current selection. "
    "It will always display what you can select next, "
    "or an exercise description if you have selected up to an exercise."
)
@click.option(
    "-force", "--force",
    help="Force display (with possible mistakes) the exercise (-series) "
    "description to display.",
    is_flag=True, default=False
)
def display(force):
    import http.client
    from dodonacli.source import get_data, pretty_print

    # Read configs in
    config = get_data.get_configs()
    settings = get_data.get_settings()

    # Start up the connection to Dodona
    connection = http.client.HTTPSConnection("dodona.be")
    headers = {
        "Content-type": "application/json",
        "Accept": "application/json",
        "Authorization": config['TOKEN']
    }

    # Display flag changes behaviour depending on the values in the
    # config-dictionary.
    if config['course_id'] is None:
        # Print available courses
        json_data = get_data.courses_data(connection, headers)
        pretty_print.print_courses_data(json_data, settings)

    elif config['serie_id'] is None:
        # Print available series
        json_data = get_data.series_data(connection, headers, config['course_id'])
        pretty_print.print_series_data(json_data, settings, force)

    elif config['exercise_id'] is None:
        # Print available exercises
        if config['serie_token'] is None:
            serie_token = ""
        else:
            serie_token = "?token=" + config['serie_token']

        json_data = get_data.exercises_data(
            connection, headers, config['serie_id'], serie_token
        )
        pretty_print.print_exercise_data(json_data, settings)

    else:
        # Print exercise-description
        settings = get_data.get_settings()
        json_data = get_data.exercise_data(
            connection, headers, config['course_id'], config['exercise_id']
        )
        pretty_print.print_exercise(json_data, config['TOKEN'], settings, force)



import os
import shutil
import subprocess
import tempfile

from .pretty_printer import custom_print as _custom_print


def check_syntax(file: str, language: str, settings: dict) -> bool:
    """
    Check the syntax of the code in the file using external tools
    :param file: Filename containing the code to check
    :param language: Programming language to determine what tool to use
    :return: Returns whether the syntax is right
    """
    def custom_print(text):
        _custom_print(text, settings)

    if language in ("bash", "sh"):
        return check_bash_syntax(file, custom_print)
    if language == "python":
        return check_python_syntax(file, custom_print)
    if language == "java":
        return check_java_syntax(file, custom_print)
    if language == "javascript":
        return check_javascript_syntax(file, custom_print)

    custom_print("Language syntax check is not implemented for: " + language)

    return True


def check_bash_syntax(file: str, custom_print) -> bool:
    try:
        # Exclude:
        # SC2001: See if you can use ${variable//search/replace} instead.
        subprocess.run(['shellcheck', '-e', 'SC2001', file], check=True)
        return True
    except subprocess.CalledProcessError as cpe:
        custom_print(f"Syntax Error in Bash script: {str(cpe)}")
        return False
    except FileNotFoundError:
        # This will only occur if shellcheck isn't installed.
        # Click will detect that the user gave an invalid file before this
        # function is called
        custom_print(
            "To check the syntax, 'shellcheck' is called with your file. "
            "It appears however, that shellcheck isn't installed on your "
            "system. \nPlease install it: https://www.shellcheck.net/"
        )
        return False
    except Exception as e:
        custom_print(
            "No idea what's going wrong, "
            "but something definitly is going wrong:"
            + str(e)
        )
        return False


def check_python_syntax(file: str, custom_print) -> bool:
    try:
        with open(file, 'r') as content:
            compile(content.read(), file, 'exec')
        return True
    except SyntaxError as se:
        custom_print(f"Syntax error in Python file:\n{se}")
    except Exception as e:
        custom_print(
            "No idea what's going wrong, "
            "but something definitly is going wrong:"
            + str(e)
        )
        return False


def check_java_syntax(file: str, custom_print) -> bool:
    temp_dir = tempfile.mkdtemp()
    try:
        directory = os.path.dirname(file)
        java_files = [
            file
            for file in os.listdir(directory)
            if file.endswith('.java') and "test" not in file.lower()
        ]
        if not java_files:
            return False  # No Java files found in the directory

        # Construct the list of Java files with their full paths
        java_files_with_paths = [
            os.path.join(directory, file)
            for file in java_files
        ]

        subprocess.run(
            ['javac', '-d', temp_dir] + java_files_with_paths,
            check=True
        )

        # Remove temporary directory
        shutil.rmtree(temp_dir, ignore_errors=True)
        return True

    except subprocess.CalledProcessError:
        # This error will happen if there was something wrong with the
        # Java syntax
        return False

    except FileNotFoundError:
        # This will only occur if javac isn't installed.
        # Click will detect that the user gave an invalid file before this
        # function is called
        custom_print(
            "To check the syntax, 'javac' is called with your file. "
            "It appears however, that the Java compiler isn't installed on "
            "your system.\nPlease install it: https://openjdk.org/install/"
        )
        return False
    except Exception as e:
        custom_print(
            "No idea what's going wrong, "
            "but something definitly is going wrong:"
            + str(e)
        )
        return False
    finally:
        shutil.rmtree(temp_dir, ignore_errors=True)


def check_javascript_syntax(file: str, custom_print) -> bool:
    try:
        subprocess.run(['node', '-c', file], check=True)
        return True
    except subprocess.CalledProcessError:
        return False
    except FileNotFoundError:
        # This will only occur if jshint isn't installed.
        # Click will detect that the user gave an invalid file before this
        # function is called
        custom_print(
            "To check the syntax, 'node' is called with your file. "
            "It appears however, that this program isn't installed on "
            "your system.\nPlease install it: "
            "https://nodejs.org/en/learn/getting-started/how-to-install-nodejs"
        )
        return False
    except Exception as e:
        custom_print(
            "No idea what's going wrong, "
            "but something definitly is going wrong:"
            + str(e)
        )
        return False


def check_haskell_syntax(file: str, custom_print) -> bool:
    try:
        # TODO: enable compiling without main function present
        subprocess.run(['ghc', '-fno-code', '-v0', file], check=True)
        return True
    except subprocess.CalledProcessError:
        return False
    except FileNotFoundError:
        custom_print(
            "To check the syntax, 'ghc' is called with your file. "
            "It appears however, that this program isn't installed on "
            "your system.\nPlease install it: https://www.haskell.org/ghcup/"
        )
        return False
    except Exception as e:
        custom_print(
            "No idea what's going wrong, "
            "but something definitly is going wrong:"
            + str(e)
        )
        return False



import click


@click.command(help="Start tutorial")
def tutorial():
    from dodonacli.source import set_data, get_data, interactive_tutorial

    config = get_data.get_configs()
    settings = get_data.get_settings()
    config = interactive_tutorial.start_tutorial(config, settings)
    set_data.dump_config(config)



def submission_data_handler(submission_data: dict, settings: dict) -> str:
    result = ""
    if submission_data['description'] == "Timeout":
        result += "\n[bold bright_red]Timout[/]"
    else:
        result += submission_tabs_handler(submission_data, settings)
    result += submission_code_annotations(submission_data)
    return result


def submission_tabs_handler(submission_data: dict, settings: dict) -> str:
    """
    Handle all tabs, group together those that were accepted and delegate the
    failed tabs
    :param submission_data: dictionary with the results of a submission
    :param settings: dictionary with settings
    :return: formatted string with info per tab
    """
    failed_tabs = []
    correct_tabs = []
    for tab in submission_data['groups']:
        if tab['badgeCount'] > 0:
            failed_tabs.append(tab)
        else:
            correct_tabs.append(tab)

    if settings['amount_feedback_tab'] == -1:
        amount_failed_display = len(failed_tabs)
    else:
        amount_failed_display = min(
            settings['amount_feedback_tab'], len(failed_tabs)
        )

    result = "[bold bright_red]Wrong tabs[/]:"
    result += ''.join(
        failed_tab_handler(failed_tab, settings)
        for failed_tab in failed_tabs[:amount_failed_display]
    )

    if len(correct_tabs) > 0:
        result += "\n\n[bold bright_green]Correct tabs:[/] "
        result += ', '.join(tab['description'] for tab in correct_tabs)

    if amount_failed_display < len(failed_tabs):
        result += "\n[bold bright_red]Other failed tabs:[/] "
        result += ', '.join(
            tab['description']
            for tab in failed_tabs[amount_failed_display:]
        )

    return result


def failed_tab_handler(tab: dict, settings: dict) -> str:
    """
    Handle failed tabs by delegating the first n failed contexts
    :param tab: dictionary with a failed tab
    :param settings: dictionary with settings
    :return: formatted string with failed tab things
    """
    result = (
        f"\n[bold bright_red] \u00B7 [/]"
        f"[u]{str(tab['description']).capitalize()} ({tab['badgeCount']}):[/]"
    )
    contexts = tab['groups']
    index = 0
    amount_failed = 0
    max_shown = settings['amount_feedback_context']

    while index < len(contexts) and amount_failed < max_shown:
        context = contexts[index]
        if context['accepted']:
            index += 1
            continue
        result += failed_context_handler(context, settings)
        amount_failed += 1
        index += 1
    return result


def failed_context_handler(context: dict, settings: dict) -> str:
    """
    Handle failed context by delegating the first n failed contexts
    :param context: dictionary with a failed context
    :param settings: dictionary with settings
    :return: formatted string with failed tab things
    """
    result = ""

    test_cases = context['groups']
    index = 0
    amount_failed = 0
    max_shown = settings['amount_feedback_testcase']

    context_content = []

    while index < len(test_cases) and amount_failed < max_shown:
        test_case = test_cases[index]

        if 'tests' in test_case:
            context_content.extend(failed_tests_handler(test_case, settings))
        else:
            if test_case['accepted']:
                emoji = '[bright_green]:heavy_check_mark:[/] '
            else:
                emoji = "[bright_red]:heavy_multiplication_x:[/] "
            context_content.append(
                emoji + test_case['description']['description'].rstrip()
            )

        if not test_case['accepted']:
            amount_failed += 1
        index += 1

    if len(context_content) > 1:
        result += f"\n   \u256D {context_content[0]}"
        result += (len(context_content[1:-1]) > 0) * "\n   \u2502 "
        result += "\n   \u2502 ".join(cc for cc in context_content[1:-1])
        result += f"\n   \u2570 {context_content[-1]}\n"
    else:
        result += "\n   - " + context_content[0] + "\n"

    return result


def failed_tests_handler(test_case: dict, settings: dict) -> list:
    """
    Handle failed test_case by displaying all tests
    :param test_case: dictionary with a failed context
    :param settings: dictionary with settings
    :return: formatted string with failed tab things
    """
    result = []
    if 'data' in test_case:
        result.append(test_case['data']['statements'].rstrip())
    elif 'description' in test_case:
        result.append(test_case['description']['description'].rstrip())

    tests = test_case['tests']
    if all(test['accepted'] for test in tests):
        result[-1] = "[bright_green]:heavy_check_mark:[/] " + result[-1]
        return result
    else:
        result[-1] = "[bright_red]:heavy_multiplication_x:[/] " + result[-1]

    index = 0
    amount_failed = 0
    max_shown = settings['amount_feedback_test']

    while index < len(tests) and amount_failed < max_shown:
        test = tests[index]
        if not test['accepted']:
            result.append("\t    Expected:\t" + test['expected'].rstrip())
            result.append("\t    Actual:  \t" + test['generated'].rstrip())
        amount_failed += 1
        index += 1

    return result


def submission_code_annotations(submission_data: dict) -> str:
    """
    Get code annotations out of the submission data
    :param submission_data: dictionary with the results of a submission
    :return: formatted string with code annotations
    """
    result = ""
    sub_annotations = submission_data['annotations']
    if 'annotations' in submission_data and len(sub_annotations) > 0:
        result = "\n\n[bold]Code annotations:[/]"
        for annotation in sub_annotations:
            result += f"\n- Row {annotation['row']}: {annotation['text']}"
    return result



def save_boilerplate(content: str, exercise_name: str, extension: str):
    import os.path
    import textwrap
    from . import pretty_console

    boiler_filename = f"{exercise_name.lower()}.{extension}"

    if os.path.isfile(boiler_filename):
        print(
            f"\nNot saving boilerplate code because '{boiler_filename}' "
            "already exists.\nBoilerplate:\n"
        )
        pretty_console.console.print(
            textwrap.indent(content.rstrip(), '\t')
        )
    else:
        print(f"\nBoilerplate code (copy in '{boiler_filename}'):\n")
        pretty_console.console.print(
            textwrap.indent(content.rstrip(), '\t')
        )
        with open(boiler_filename, "w") as boilerplate_file:
            boilerplate_file.write(content)
    pass



import click


@click.command(
    help="Deselect default last selected thing. "
    "Can be used with an argument to deselect everything with "
    "'all' or 'top', or deselect 1, 2 or 3 levels."
)
@click.argument(
    'amount', default='1',
    type=click.Choice(['all', 'top', '1', '2', '3'], case_sensitive=False)
)
def up(amount):
    from dodonacli.source import set_data, get_data
    from dodonacli.source.pretty_printer import custom_print

    # Read configs in
    config = get_data.get_configs()
    settings = get_data.get_settings()

    # Click handles all the edge cases, we can just assume it's a valid string
    if amount in ('all', 'top'):
        keys_to_remove = (
            "course_id", "course_name",
            "serie_id", "serie_name",
            "exercise_id", "exercise_name",
            "serie_token", "programming_language"
        )
        for e in keys_to_remove:
            config[e] = None

        custom_print("Deselected everything.", settings)

    else:
        levels = int(amount)
        string_to_print = ""
        for _ in range(levels):
            if config.get('exercise_id') is not None:
                config['exercise_id'] = None
                config['exercise_name'] = None
                config['programming_language'] = None
                string_to_print += "Deselected exercise.\n"
            elif config.get('serie_id') is not None:
                config['serie_id'] = None
                config['serie_name'] = None
                config['serie_token'] = None
                string_to_print += "Deselected series.\n"
            elif config.get('course_id') is not None:
                config['course_id'] = None
                config['course_name'] = None
                string_to_print += "Deselected course.\n"
            else:
                string_to_print += "Nothing selected.\n"
                break
        custom_print(string_to_print, settings, end='')

    set_data.dump_config(config)



import click


@click.command(
    help="Post a solution-file to Dodona. "
    "The file has to be in your current working directory, "
    "and this only works if there is a selected exercise. "
    "DodonaCLI will give a try at rendering a small part of the feedback-table"
    " so you can continue working from the terminal, this isn't guarenteed to "
    "provide enough info, but it's doing its best."
)
@click.option(
    "-l", "--use-link",
    help="Post your solutionfile to the link at the first line of your "
    "solutionfile. "
    "This is inspired by plugins for editors as VSCode for Dodona.",
    is_flag=True, default=False
)
@click.option(
    "-c", "--check",
    help="Check the file you provided if the syntax is valid for the "
    "programming language associated with the exercise. "
    "Currently supported languages: bash, python, java, " "javascript. "
    "For Java files, it will use javac to compile all *.java files that "
    "don't have 'test' in them.",
    is_flag=True, default=False
)
@click.argument(
    'file',
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, resolve_path=True
    )
)
def post(file, use_link, check):
    import http.client
    from dodonacli.source import set_data, get_data, syntax_checker

    # Read configs in
    config = get_data.get_configs()
    settings = get_data.get_settings()

    # Quick check if combination of selected exercise and -l flag are valid
    if not use_link and not config['exercise_id']:
        print(
            '\n' * settings['new_lines_above']
            + "No exercise selected! If you want to use a link at the top of "
            + "your file, use the -l flag."
            + '\n' * settings['new_lines_below']
        )
        return

    # Must be filled in:
    content = None
    course_id = None
    exercise_id = None

    # Check for the link at the top of the file
    if use_link:
        # TODO: move this link extraction to different function (diff. file?)
        with open(file, 'r') as solutionfile:
            link = solutionfile.readline()

            # If file starts with hashbang, look at 2nd line
            hashbang = None
            if link[:2] == '#!':
                hashbang = link
                link = solutionfile.readline()

            # Check if valid link to try to post to
            link_index = link.find("https://dodona.be/")
            if link_index < 0:
                print(
                    '\n' * settings['new_lines_above']
                    + "No valid link found on the first line of your file. "
                      "Please confirm again.\n"
                      "A valid link starts with 'https://dodona.be/'"
                    + '\n' * settings['new_lines_below']
                )
                return
            link = link[link_index:]

            # Take out the id's we need
            if link.find("/courses/") != -1:
                course_id_index_start = link.find("/courses/")
                course_id_index_start += len("/courses/")

                course_id_index_stop = link.find("/", course_id_index_start)

                course_id = link[course_id_index_start:course_id_index_stop]
            else:
                course_id = None

            exercise_id_index_start = link.find("/activities/")
            exercise_id_index_start += len("/activities/")

            exercise_id_index_stop = link.find("/", exercise_id_index_start)

            exercise_id = link[exercise_id_index_start:exercise_id_index_stop]

            # We can read the content of the file now as the file-pointer is
            # already at the 2nd line
            if hashbang:
                content = hashbang + solutionfile.read()
            else:
                content = solutionfile.read()

    else:
        # Syntax check only available without link at top of file
        if check:
            syntax_ok = syntax_checker.check_syntax(
                file, config['programming_language'], settings
            )
            if not syntax_ok:
                return

        with open(file, 'r') as solutionfile:
            content = solutionfile.read()

        course_id = config['course_id']
        exercise_id = config['exercise_id']

    # Make sure the amount of newlines is exactly 1 to make Dodona's linters
    # happy
    content = content.rstrip() + "\n"

    # Start up the connection to Dodona
    connection = http.client.HTTPSConnection("dodona.be")
    headers = {
        "Content-type": "application/json",
        "Accept": "application/json",
        "Authorization": config['TOKEN']
    }

    # Post exercise to Dodona
    set_data.post_solution(
        content, connection, headers, course_id, exercise_id, settings
    )
